'use strict';

import { useEffect, useRef, useCallback } from 'react';
import { findNodeHandle } from 'react-native';
import { useEvent } from './useEvent';
import { useSharedValue } from './useSharedValue';
import { isWeb } from '../PlatformChecker';
const IS_WEB = isWeb();

/**
 * Lets you synchronously get the current offset of a `ScrollView`.
 *
 * @param animatedRef - An [animated ref](https://docs.swmansion.com/react-native-reanimated/docs/core/useAnimatedRef) attached to an Animated.ScrollView component.
 * @returns A shared value which holds the current offset of the `ScrollView`.
 * @see https://docs.swmansion.com/react-native-reanimated/docs/scroll/useScrollViewOffset
 */
export const useScrollViewOffset = IS_WEB ? useScrollViewOffsetJS : useScrollViewOffsetNative;
function useScrollViewOffsetJS(animatedRef, initialRef) {
  const offsetRef = useRef(
  // eslint-disable-next-line react-hooks/rules-of-hooks
  initialRef !== undefined ? initialRef : useSharedValue(0));
  const scrollRef = useRef(null);
  const eventHandler = useCallback(() => {
    'worklet';

    const element = animatedRef.current;
    // scrollLeft is the X axis scrolled offset, works properly also with RTL layout
    offsetRef.current.value = element.scrollLeft === 0 ? element.scrollTop : element.scrollLeft;
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [animatedRef, animatedRef.current]);
  useEffect(() => {
    // We need to make sure that listener for old animatedRef value is removed
    if (scrollRef.current !== null) {
      scrollRef.current.removeEventListener('scroll', eventHandler);
    }
    scrollRef.current = animatedRef.current;
    const element = animatedRef.current;
    element.addEventListener('scroll', eventHandler);
    return () => {
      element.removeEventListener('scroll', eventHandler);
    };
    // React here has a problem with `animatedRef.current` since a Ref .current
    // field shouldn't be used as a dependency. However, in this case we have
    // to do it this way.
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [animatedRef, animatedRef.current, eventHandler]);
  return offsetRef.current;
}
const scrollNativeEventNames = ['onScroll', 'onScrollBeginDrag', 'onScrollEndDrag', 'onMomentumScrollBegin', 'onMomentumScrollEnd'];
function useScrollViewOffsetNative(animatedRef, initialRef) {
  const offsetRef = useRef(
  // eslint-disable-next-line react-hooks/rules-of-hooks
  initialRef !== undefined ? initialRef : useSharedValue(0));
  const eventHandler = useEvent(event => {
    'worklet';

    offsetRef.current.value = event.contentOffset.x === 0 ? event.contentOffset.y : event.contentOffset.x;
  }, scrollNativeEventNames
  // Read https://github.com/software-mansion/react-native-reanimated/pull/5056
  // for more information about this cast.
  );

  useEffect(() => {
    const component = animatedRef.current;
    const viewTag = IS_WEB ? component : findNodeHandle(component);
    eventHandler.workletEventHandler.registerForEvents(viewTag);
    return () => {
      var _eventHandler$worklet;
      (_eventHandler$worklet = eventHandler.workletEventHandler) === null || _eventHandler$worklet === void 0 ? void 0 : _eventHandler$worklet.unregisterFromEvents();
    };
    // React here has a problem with `animatedRef.current` since a Ref .current
    // field shouldn't be used as a dependency. However, in this case we have
    // to do it this way.
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [animatedRef, animatedRef.current, eventHandler]);
  return offsetRef.current;
}
//# sourceMappingURL=useScrollViewOffset.js.map